import re
from typing import Type, TypeVar, Any

import venusian
from pyramid.config import Configurator
from zope.interface import Interface
from zope.interface.interface import InterfaceClass
from functools import update_wrapper
from pyramid_services import _resolve_iface
import warnings


__version__ = "0.4.dev0"


_to_underscores = re.compile("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")


class reify_attr(object):
    """
    reify_attr is like pyramid reify, but instead of getting the name of the
    attribute from the decorated method, it uses the name of actual attribute,
    by finding it on the class in Python <=3.5, and using the ``__set_name__``
    on Python 3.6.
    """

    names = None

    def __init__(self, wrapped):
        self.wrapped = wrapped
        update_wrapper(self, wrapped)

    def __get__(self, inst, objtype=None):
        if inst is None:
            return self

        if self.names is None:
            raise TypeError(
                f"reify_attr decorating {self.wrapped} not bound to a named attribute!"
            )

        val = self.wrapped(inst)
        for name in self.names:
            setattr(inst, name, val)

        return val

    def __set_name__(self, owner, name):
        if self.names is None:
            self.names = [name]

        else:
            self.names.append(name)


def _underscore(name):
    return _to_underscores.sub(r"_\1", name).lower()


_is_iface_name = re.compile("^I[A-Z].*")


class ServiceRegistry(object):
    def __init__(self):
        self.__services__ = []

    def _register_service(self, instance, interface):
        self.__services__.append((instance, interface))
        name = interface.__name__
        if interface is _resolve_iface(interface) and _is_iface_name.match(name):
            name = name[1:]

        setattr(self, _underscore(name), instance)


def get_service_registry(registry):
    if not hasattr(registry, "services"):
        registry.services = ServiceRegistry()

    return registry.services


def register_di_service(
    config: Configurator,
    service_factory,
    *,
    scope,
    interface=Interface,
    name="",
    context_iface=Interface,
):
    registry = config.registry
    if scope == "global":
        warnings.warn(
            "The use of 'global' as scope parameter is deprecated. Use 'application' instead",
            DeprecationWarning,
        )
        scope = "application"

    if scope == "application":
        # register only once
        real_interface = _resolve_iface(interface)
        if registry.queryUtility(real_interface, name=name) is None:
            ob_instance = service_factory(registry=registry)
            get_service_registry(registry)._register_service(ob_instance, interface)

            registry.registerUtility(ob_instance, real_interface, name=name)

            config.register_service(
                service=ob_instance, iface=interface, context=context_iface, name=name
            )

        else:
            warnings.warn(
                f"Double registration of the same service {interface}"
                f" with name {name!r} attempted"
            )

    else:
        # noinspection PyUnusedLocal
        def wrapped_factory(context, request):
            return service_factory(request=request)

        config.register_service_factory(
            wrapped_factory, interface, context_iface, name=name
        )


_NOT_SET = object()


def service(interface=None, *, name="", context_iface=Interface, scope=_NOT_SET):

    if scope not in {"global", "application", "request", _NOT_SET}:
        raise ValueError(
            f"Invalid scope {scope}, must be either 'application' or 'request'"
        )

    if scope == "global":
        warnings.warn(
            "The use of 'global' as scope parameter is deprecated. Use 'application' instead",
            DeprecationWarning,
        )

    service_name = name

    def service_decorator(wrapped):
        # noinspection PyUnusedLocal,PyShadowingNames
        def callback(scanner, name, ob):
            config = scanner.config

            iface = interface
            if iface is None:
                if service_name:
                    iface = Interface
                else:
                    iface = ob

            actual_scope = scope
            if actual_scope is _NOT_SET:
                actual_scope = getattr(ob, "__pyramid_di_scope__", None)
                if actual_scope is None:
                    raise TypeError(f"Cannot infer service scope for {ob}")

                if actual_scope not in {"application", "request"}:
                    raise TypeError(
                        f"{ob} has invalid scope {actual_scope}, must be either 'application' or 'request'"
                    )

            config.register_di_service(
                ob,
                name=service_name,
                interface=iface,
                context_iface=context_iface,
                scope=actual_scope,
            )

        venusian.attach(wrapped, callback, category="pyramid_di.service")
        return wrapped

    return service_decorator


T = TypeVar("T", bound=object)


def autowired(interface: Type[T] = Interface, name: str = "") -> T:
    def getter(self) -> T:
        if hasattr(self, "request"):
            context = getattr(self.request, "context", None)
            return self.request.find_service(interface, context, name)

        return self.registry.getUtility(_resolve_iface(interface), name)

    return reify_attr(getter)


class ApplicationScopedBaseService(object):
    """
    A convenience base class for application-scoped services.
    """

    __pyramid_di_scope__ = "application"

    registry: "pyramid.registry.Registry"

    def __init__(self, *, registry: "pyramid.registry.Registry", **kw):
        self.registry = registry
        super(ApplicationScopedBaseService, self).__init__(**kw)


class BaseService(ApplicationScopedBaseService):
    def __init__(self, **kw):
        warnings.warn(
            "BaseService has been renamed to ApplicationScopedBaseService",
            DeprecationWarning,
            2,
        )
        super(BaseService, self).__init__(**kw)

    def __init_subclass__(self):
        warnings.warn(
            "BaseService has been renamed to ApplicationScopedBaseService",
            DeprecationWarning,
            2,
        )


class RequestScopedBaseService(object):
    """
    A convenience class for request-scoped services.
    """

    __pyramid_di_scope__ = "request"

    request: "pyramid.request.Request"
    registry: "pyramid.registry.Registry"
    context: Any

    def __init__(self, *, request: "pyramid.request.Request", **kw):
        self.request = request
        self.context = getattr(request, "context", None)
        self.registry = request.registry
        super(RequestScopedBaseService, self).__init__(**kw)


def scan_services(config, *a, **kw):
    kw["categories"] = ("pyramid_di.service",)
    return config.scan(*a, **kw)


def includeme(config):
    config.include("pyramid_services")
    config.add_directive("scan_services", scan_services)
    config.add_directive("register_di_service", register_di_service)
    get_service_registry(config.registry)



import pytest
from pyramid.config import Configurator
from pyramid.request import Request
from pyramid.router import Router
import random
from pyramid.registry import Registry

from pyramid_di import BaseService, RequestScopedBaseService, autowired, service, reify_attr, ApplicationScopedBaseService

from zope.interface import Interface


class IInterfacedService(Interface):
    def something():
        ...


@service(interface=IInterfacedService)
class InterfacedService(ApplicationScopedBaseService):
    def something(self):
        return 'something'


@service()
class GlobalService(ApplicationScopedBaseService):
    def baz(self):
        return 'global'


@service()
class ServiceTwo(ApplicationScopedBaseService):
    global_service_dependency = autowired(GlobalService)
    interfaced_service = autowired(IInterfacedService)

    def bar(self):
        if (self.global_service_dependency.baz() == 'global' and
            self.interfaced_service.something() == 'something'):
            return 'ServiceTwo.bar'


@service(scope='request', name='name_only')
class NamedService(RequestScopedBaseService):
    def foo(self):
        return 'named'


@service(scope='request')
class ServiceOne(RequestScopedBaseService):
    dependency = autowired(ServiceTwo)
    named_dependency = autowired(name='name_only')

    def foo(self):
        return self.dependency.bar()


@pytest.fixture()
def config():
    with Configurator() as config:
        config.include('pyramid_di')
        config.scan_services(__name__)

    return config

@pytest.fixture()
def pyramid_app(config) -> Router:
    return config.make_wsgi_app()


@pytest.fixture()
def pyramid_request(pyramid_app) -> Request:
    wsgi_env = {}
    with pyramid_app.request_context(wsgi_env) as request:
        yield request


class ReifyAttrTest:
    @reify_attr
    def foo(self):
        return 42

    @reify_attr
    def ham(self):
        return random.random()

    spam = ham

def test_service_decorator():
    with pytest.raises(ValueError, match='.*Invalid scope.*'):
        @service(scope='something else')
        class Test:
            pass

    with pytest.warns(DeprecationWarning, match='.*global.*'):
        @service(scope='global')
        class Test:
            pass

    with pytest.warns(DeprecationWarning, match='.*renamed.*'):
        class Test(BaseService):
            pass

    with pytest.warns(DeprecationWarning, match='.*renamed.*'):
        Test(registry=Registry())


def test_reify_attr():
    assert type(ReifyAttrTest.foo) is reify_attr
    rat = ReifyAttrTest()
    assert rat.foo == 42

    ReifyAttrTest.bar = reify_attr(lambda self: 666)
    with pytest.raises(TypeError, match='.*not bound to a named attribute.*') as excinfo:
        rat.bar()

    assert rat.ham is rat.spam


def test_everything(pyramid_request):
    service_one = pyramid_request.find_service(ServiceOne)
    assert service_one.foo() == 'ServiceTwo.bar'
    assert service_one.named_dependency.foo() == 'named'


def test_reregistration_raises_warning(config):
    with pytest.warns(UserWarning, match='.*Double registration of the same service.*'):
        config.register_di_service(InterfacedService, scope='global')



import os, sys

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.md")).read()
CHANGES = open(os.path.join(here, "CHANGES.md")).read()

requires = ["pyramid>=1.7,<2.0", "pyramid_services>=1.1,<2.0"]
dev_requires = ["pytest"]

setup(
    name="pyramid_di",
    version="0.4.0",
    description="Dependency injection stuff for Pyramid",
    long_description=README + "\n\n" + CHANGES,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Pyramid",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Python Software Foundation License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    author="Antti Haapala",
    author_email="antti.haapala@anttipatterns.com",
    url="http://github.com/tetframework/pyramid_di",
    keywords="web wsgi bfg pylons pyramid",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    extras_require={"dev": dev_requires},
)
